"""This package enables you receive BMRS data as json instead of the default XML"""
__version__ = "2021.07.01"
# this new verison sleeps for 30 seconds in case of any error/exception and kees retrying infinitely

from time import sleep, time
import stomp
import xmltodict
import json
import socket


def connect_and_subscribe(conn, api_key, client_id):
    acknowledgement_mode = "client-individual"
    headers = {
        "activemq.subcriptionName": client_id,
        "activemq.subscriptionName": client_id,
    }
    conn.connect(api_key, api_key, wait=True, headers={"client-id": client_id})
    conn.subscribe(
        destination="/topic/bmrsTopic",
        ack=acknowledgement_mode,
        id=client_id,
        headers=headers,
    )


def get_hostname():
    """Returns hostname of  to be used as client id for connection"""
    return socket.getfqdn()


class MyListener(stomp.ConnectionListener):
    """This is a listener class that listens for new messages using the STOMP protocol"""

    def __init__(self, conn, listener, api_key, client_id):
        self.listener = listener
        self.conn = conn
        self.api_key = api_key
        self.client_id = client_id
        self.acknowledgement_mode = "client-individual"

    def on_error(self, headers, message):
        print(f'on_error! : "{headers}"')
        self.reconnect()

    def on_message(self, headers, message):
        message = xmltodict.parse(message)
        for key, value in headers.items():
            message[key] = value
            # print('header: key %s , value %s' %(key, value))
        message = json.loads(json.dumps(message))
        self.listener(message)
        # acknowledge message
        ack = headers["ack"]
        self.conn.ack(ack)

    def on_disconnected(self):
        print("disconnected")
        self.conn.disconnect()
        self.reconnect()

    def on_heartbeat_timeout(self):
        print("heartbeat_timeout")
        self.reconnect()

    def reconnect(self):
        print("attemptig reconnect")
        connect_and_subscribe(self.conn, self.api_key, self.client_id)


def connect_to_api(api_key, listener, client_id="", port=61613):
    """ Connect to the BMRS API. This function will get the XML data and convert it to JSON instantaneously!
    `api_key`:str is your API key that can be got by following [Guide](https://www.elexon.co.uk/documents/training-guidance/bsc-guidance-notes/bmrs-api-and-data-push-user-guide-2/).
    `client_id`:str is your Client ID that can be got together with the APIkey from the above step.
    `listener`:func is your custom function that receives & handles messages returned from the API. See a sample of this in our sample file sample_client.py on github.
    """
    # generate client id if not supplied
    if not client_id:
        client_id = get_hostname()
    print('client_id', client_id)
    # connect to API using stomp
    conn = stomp.Connection12(
        host_and_ports=[("api.bmreports.com", port)], use_ssl=True
    )
    conn.set_listener("", MyListener(conn, listener, api_key, client_id))

    while True:
        ## parts that can be moved to inside the loop
        connect_and_subscribe(conn, api_key, client_id)
        # check for new messages after every x seconds
        while conn.is_connected():
            sleep(1)
        conn.disconnect()



"""
Module to Connect to the Elexon BMRS service and return REMIT notifications among others
"""
import stomp
import time
from utils.listener import StompListener
from utils.custom_logger import CustomLogger
import socket


def get_hostname():
    """
    Returns hostname of computer
    """
    return socket.gethostname()


class DurableStompConnection:
    """
    Handles the creation of the durable STOMP connection and the processing of the data
    """

    def __init__(
        self,
        api_key,
        bmrs_message_processor,
        logger=CustomLogger(),
        client_id=get_hostname(),
        message_types=[],
        bmrs_host="api.bmreports.com",
        port=61613,
        durable=True,
        poll_interval=1,
        xml_to_json=True,
    ):
        self.__logger = logger
        self.__durable = durable
        self.__poll_interval = poll_interval
        self.__bmrs_message_processor = bmrs_message_processor
        self.__api_key = api_key
        self.__client_id = client_id
        self.__message_types = message_types
        self.__xml_to_json = xml_to_json
        self.__connection = stomp.Connection12(
            host_and_ports=[(bmrs_host, port)], use_ssl=True
        )

    def __diconnection_handler(self, message):
        self.__logger.error(f'ERROR! : "{message}"')
        self.__restart()

    def __generate_jms_type_selector_string(self):
        """
        Generates a selector string that allows all the message_types
        passed to be listened to
        """
        selector_substrings = [
            "JMSType='{}'".format(message_type) for message_type in self.__message_types
        ]
        return " OR ".join(selector_substrings)

    def __acknowledge_message(self, ack):
        """
        Sends an ACK frame to the server to acknowledge receipt
        http://jasonrbriggs.github.io/stomp.py/api.html#acks-and-nacks
        """
        self.__connection.ack(ack)

    def __process_bmrs_message(self, error, message_dict, headers={}):
        """
        Processes parsed bmrs messages
        """
        if error is not None:
            self.__logger.error("Error thrown {}".format(error))

        try:
            ack = headers["ack"]
            self.__acknowledge_message(ack)

        except KeyError as exception:
            self.__logger.error("No Message Id received")
            raise exception

        if message_dict is not None:
            self.__bmrs_message_processor(message_dict)
            self.__logger.info("NEW MESSAGE RECEIVED \n\n")

    def __restart(self):
        """
        Restarts listening for messages but must first clean up
        """
        try:
            self.disconnect()
        except stomp.exception.NotConnectedException:
            pass
        self.start()

    def disconnect(self):
        """
        Unsubscribes and Disconnects the listener from the server
        """
        self.__connection.unsubscribe(self.__client_id)
        return self.__connection.disconnect()

    def __connect(self):
        """
        Connects to the remote broker
        For durable connections:
            - parameter 'ack' = 'client-individual' for the connection
        """
        self.__connection.connect(
            self.__api_key,
            self.__api_key,
            wait=True,
            headers={"client-id": self.__client_id},
        )

    def __subscribe(self):
        """
        Creates a connection and subscription: durable by default
        https://stomp.github.io/stomp-specification-1.2.html#SUBSCRIBE
        https://activemq.apache.org/stomp
        For durable connections:
            - parameter 'ack' = 'client-individual' for the connection
            - header 'activemq.subcriptionName' of subscription = header 'client-id'
            for the connection in ActiveMQ versions prior 5.7.0
            (note: 'subcriptionName' not 'subscriptionName')
            - header 'activemq.subscriptionName' of subscription = header 'client-id'
            for the connection in ActiveMQ versions after 5.6.0
            (note: 'subscriptionName' not 'subcriptionName')
        """
        acknowledgement_mode = "client-individual" if self.__durable else "auto"
        headers = {
            "activemq.subcriptionName": self.__client_id,
            "activemq.subscriptionName": self.__client_id,
        }

        if len(self.__message_types) > 0:
            headers["selector"] = self.__generate_jms_type_selector_string()

        self.__connection.subscribe(
            destination="/topic/bmrsTopic",
            ack=acknowledgement_mode,
            id=self.__client_id,
            headers=headers,
        )

    def start(self):
        """
        Initializes the STOMP client and connects to BMRS and listens
        """
        self.__connection.set_listener(
            "",
            StompListener(
                self.__process_bmrs_message,
                disconnection_handler=self.__diconnection_handler,
                xml_to_json=self.__xml_to_json,
            ),
        )
        self.__connection.start()
        self.__connect()
        self.__subscribe()

        while self.__connection.is_connected():
            try:
                time.sleep(self.__poll_interval)
            except KeyboardInterrupt as exp:
                print(exp)
                exit(0)



'''This could be your example consumer file from which you will receive and process the API responses'''
import bmrs


def on_message(json_response):
    '''Receive a JSON response and do stuff with it'''
    print(list(json_response.keys()))
    # print(json_response)


# connect to the bmrs API using your credentials
bmrs.connect_to_api(
        api_key = 'YOUR API KEY HERE',
        client_id = 'YOUR_CLIENT_ID_HERE',
        listener=on_message
    )



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

install_requires = \
['docopt >=0.6.2', 'stomp.py >=4.1.22', 'xmltodict >=0.12.0']

setup(name='bmrs',
      version='2021.7.1',
      description='This package enables you receive BMRS data as json instead of the default XML',
      author='Edison Abahurire',
      author_email='abahedison@gmail.com',
      url='https://github.com/SimiCode/bmrs-json-api',
      py_modules=['bmrs'],
      install_requires=install_requires,
      python_requires='>=3.6',
     )
